function metadata = checkClass(obj,level,forceType) %#ok<*CHARTEN,*STREMP,*AGROW,*JAPIMATHWORKS>
% checkClass - inspect a Java, Matlab (MCOS), .NET, Python, C++ (CLib), or COM (ActiveX)
% object/classname and display its interfaces, super-classes, sub-classes,
% methods, properties, events, enumerations and annotations (where available)
%
% metadata = checkClass(...) returns a meta.class or meta.package object that
% encapsulates the input object.
%
% This utility complements the more detailed UIINSPECT utility (on the File
% Exchange) by being Command-Prompt based and also by highlighting the object
% components that were modified compared to its parent superclass.
%
% Syntax:
%    checkClass(objectReference)
%    checkClass(className)
%    checkClass(..., level) % default level=1  % Note: changed default in v1.32!
%    metadata = checkClass(...)
%
% Examples:
%    % Java objects/classes
%    checkClass('java.lang.String')
%    checkClass(javax.swing.JButton)
%    jButton=javax.swing.JButton('Click me!'); checkClass(jButton);
%    checkClass(com.mathworks.mde.desk.MLDesktop.getInstance)
%
%    % Matlab (MCOS) objects/classses
%    checkClass(gcf)
%    checkClass('matlab.ui.Figure')
%    checkClass  matlab.ui.Figure   % equivalent to above
%    checkClass  matlab.ui.control  % package name (contains classes, sub-packages)
%
%    % .NET objects/classes
%    checkClass('NET.Assembly')
%    checkClass(System.Diagnostics.Process.GetCurrentProcess)
%
%    % Python objects/classes/modules
%    checkClass('py.inspect')
%    checkClass(py.importlib.import_module('calendar'))
%
% Bugs and suggestions:
%    Please send to Yair Altman (altmany at gmail dot com)
%
% Technical description:
%    http://UndocumentedMatlab.com/blog/checkclass
%
% See also:
%    methods, methodsview, events, inspect, uiinspect (on the File Exchange)

% Release history:
%    1.0  2009-04-13: initial version
%    1.1  2009-04-14: added classes, constructors, annotations
%    1.2  2010-01-31: displayed modified methods
%    1.3  2010-01-31: fixed displayed class names
%    1.4  2010-01-31: fixed duplicate interfaces list
%    1.5  2010-01-31: enabled classname input arg
%    1.6  2010-03-14: prettier display, hyperlinked classes, error checking
%    1.7  2010-03-15: displayed static field values; fixed minor bug; displayed missing/extra constructors
%    1.8  2010-05-07: linked to checkClass (not uiinspect); fixed some edge cases; displayed non-ML superclass; displayed class modifiers
%    1.9  2010-06-16: fixed problem when directly specifying requested superclass level
%    1.10 2011-01-02: fixed static fields value display; fixed minor bug with non-derived class
%    1.11 2011-01-02: enabled checking Matlab-wrapped (javahandle_withcallbacks) class handles
%    1.12 2011-01-31: displayed function return values & qualifiers; fixed dot-notation internal classes
%    1.13 2011-02-13: fixed edge cases of no methods and problematic classnames (e.g., '$proxy')
%    1.14 2011-03-17: added hyperlinks to user-generated subclasses (myclass$subclass)
%    1.15 2011-05-30: fixed display of array of string values
%    1.16 2011-07-06: fixed hyperlinking regexp
%    1.17 2013-01-14: fixed edge case
%    1.18 2015-01-12: handled missing/string input; handled MCOS objects
%    1.19 2015-07-15: fixed java annotations
%    1.20 2017-01-19: added display of class origin information
%    1.21 2017-07-07: fixed report for some .Net classes that are reported as A*B*C
%    1.22 2017-09-12: handled a few additional edge-cases
%    1.23 2017-09-13: attribution of whereisjavaclassloadingfrom created by Andrew Janke
%    1.24 2017-12-26: handled several edge-cases of hard-to-load Java classes
%    1.25 2018-12-14: improved the display of multiple super-classes
%    1.26 2019-06-10: improved display of class hierarchies; fixes for MCOS classes
%    1.27 2019-11-27: fixed a couple of edge cases
%    1.28 2020-01-07: handled case of javahandle* objects
%    1.29 2020-09-16: improved display of Java Enums and Java classloader info
%    1.30 2021-03-12: extended property information (type, validations, default value); avoid field lengths warnings
%    1.31 2021-11-19: improved display of class description, no-arg methods, DefaultValue etc.
%    1.32 2021-11-21: added MCOS package & origin info; added javadoc link for Java classes; many fixes to superclasses, methods, description, hyperlinks; changed default level
%    1.33 2021-12-05: added package-level function info; added enumeration underlying values; added MS doc link & assembly info for .NET clases; hyperlinked code origin; return class metadata; several small fixes
%    1.34 2021-12-16: fixed code origin of packages that have no classes; improved display of function/method args in certain cases; fixed properties PartialMatchPriority display; added underlying MCOS enum values; fixed case of duplicate methods
%    1.35 2022-03-13: minor fixes to multi-line method/property descriptions (where available); ignored description changes when comparing class to superclass; merged GetAccess+SetAccess modifiers where possible; de-hyperlinked i.e.,e.g.
%    1.36 2022-04-01: improved display of MCOS property meta-data and enumeration values
%    1.37 2022-05-17: improved Python support; clarified display of redefined methods; removed display of missing superclass constructors; fixed some cases of enumerated values
%    1.38 2022-05-17: displayed methods descriptions & attributes where available; distinguished between Python package/module; minor display improvements
%    1.39 2022-08-31: display newlines in main description (unlike in methods/properties/events); display dynamic Python properties; strip irrelevant content from descriptive comments; fixed display of array DefaultValue; added separate display of MCOS constants
%    1.40 2023-01-03: fixed display of properties in abstract classes; fixed display of default empty char array value; fixed display of classes with no properties

% License to use and modify this code is granted freely to all interested, as long as the original author is
% referenced and attributed as such. The original author maintains the right to be solely associated with this work.

% Programmed and Copyright by Yair M. Altman: altmany(at)gmail.com

  if nargout, metadata = []; end
  if nargin < 1,  help(mfilename); return;  end
  if nargin < 3,  forceType = 'class';  end

  origObj = obj;
  try
      % handle Matlab-wrapped (javahandle_withcallbacks) class handles
      if strncmp(class(obj),'javahandle_',12)
          obj = java(obj);
      end
  catch
  end
  %if isempty(obj), obj = class(origObj); end
  if isjava(obj)
      thisClass = obj.getClass;
  elseif ischar(obj)
      %disp 'Not a Java object';  return;
      try
          thisClass = loadClass(obj);
      catch
          try
              % One more attempt - maybe the last sub-segment is an internal class:
              obj = regexprep(obj,'\.([^.]+$)','\$$1');
              thisClass = loadClass(obj);
          catch
              thisClass = java.lang.String(obj);
          end
      end
  elseif isempty(obj) || ~isobject(obj)
      %disp 'Uncheckable object';
      %return;
      obj = class(origObj);
      if strncmp(obj,'javahandle',10)
          try
              obj = origObj.java;
              thisClass = obj.getClass;
          catch
          end
      end
  end

  linkTarget = mfilename; %'uiinspect';
  hasUiinspect = ~isempty(which('uiinspect'));
  if ~hasUiinspect
      warning('YMA:checkClass:noUIINSPECT','UIINSPECT utility was not found. Download from <a href="http://www.mathworks.com/matlabcentral/fileexchange/17935">here</a>');
      linkTarget = mfilename;
  end

  if nargin < 2
      level = 1;  % Note: default was =Inf until v1.31, =1 since v1.32 !!!
  elseif ischar(level)
      level = str2double(level);
  end
  if numel(level)~=1 || ~isnumeric(level) || isnan(level)
      error('YMA:checkClass:badLevel','Level argument must be a number from 0-Inf');
  end
  level = max(level,1);

  isClass = true;
  objType = 'Class';
  dotNetVer = '';
  try
      isMcos = false;
      isDotNET = false;
      thisClassName = char(thisClass.getName);
      isDotNET = strncmp(thisClassName,'NET.',4) || strncmp(thisClassName,'System.',7);
  catch
      try
          % Try to parse as an MCOS object/classname
          if ischar(origObj)
              isDotNET = strncmp(origObj,'NET.',4) || strncmp(origObj,'System.',7);
              forceType = lower(forceType);
              thisClass = meta.(forceType).fromName(origObj);
              if isempty(thisClass)
                  thisClass = meta.class.fromName(origObj);
              elseif ~strcmpi(forceType,'class')
                  objType = 'Package';
                  isClass = false;
              end
              if isempty(thisClass) && any(origObj=='<') %  && isDotNET
                  parts = strsplit(origObj,'<');
                  parts{2} = strrep(parts{2},'.','*');
                  thisClass = meta.class.fromName(strjoin(parts,'<'));
              end
              if isempty(thisClass)
                  thisClass = meta.package.fromName(origObj);
                  if ~isempty(thisClass)
                      objType = 'Package';
                      isClass = false;
                  end
              end
          else
              thisClass = metaclass(origObj);
          end
          thisClassName = thisClass.Name;
          isMcos = true;
          isDotNET = strncmp(thisClassName,'NET.',4) || strncmp(thisClassName,'System.',7);
          if ~isempty(origObj), obj = origObj; end
      catch
          try
              % Last-ditch attempt to parse the input
              obj = evalin('caller',origObj);
              checkClass(obj,level);
              return  % successful - hurray!
          catch
              try
                  obj2 = strtok(which(origObj));  % 'Transport' => 'javax.mail.Transport'
                  if isempty(obj2)
                      s = dbstack;
                      obj2 = strtok(which(origObj,'in',s(2).name));
                  end
                  if isempty(obj2) || isequal(obj2,origObj), error('oops!'); end
                  checkClass(obj2,level);
                  return  % successful - hurray!
              catch
                  try
                      try objName = char(origObj); catch, objName = char(obj); end
                      str = ['<' strrep(objName,'\','\\') '>'];
                  catch
                      str = 'input object';
                  end
                  msgStr = ['Cannot process ' str ' - possibly not a class object/name, or perhaps a built-in type'];
                  error('YMA:checkClass:badInput',msgStr);
              end
          end
      end
  end
  
  % Check if it's a python object
  isPython = strncmp(thisClassName,'py.',3);
  if isPython && strcmp(thisClassName,'py.module') && ~ischar(origObj)
      [unused,rest] = strtok(strrep(char(origObj),'''','')); %#ok<ASGLU>
      pyModuleName = ['py.' strtok(strtrim(rest))];
      checkClass(pyModuleName, level);
      return
  end
  if isPython, objType = strrep(objType,'Package','Module'); end

  % Return the metadata, if requested
  if nargout
      metadata = thisClass;
  end

  % Display the class name and modifiers
  displayHeader();

  % Display package information
  getContainingPackage(thisClass, false);

  % Display superclass name(s)
  if isClass
      [superClasses,superClassNames] = getSuperClasses(thisClass,level);
      superclassFoundFlag = ~isequal(superClasses,thisClass) && ~isempty(superClasses);
  else
      %{
      superClasses = '';
      superClassNames = '';
      superclassFoundFlag = false;
      %}
      return
  end
  numSuperClasses = numel(superClasses);
  if numSuperClasses <= 1
      unifiedSuperNames = char(superClassNames);  % char() handles cell array
  elseif numSuperClasses < 4
      unifiedSuperNames = strjoin(superClassNames,',');
  else
      unifiedSuperNames = sprintf('%d superclasses',numSuperClasses);
  end
  unifiedSuperNames = regexprep(unifiedSuperNames,'[^,]*\.',''); %short class names
  if ~superclassFoundFlag, superClassNames = {}; end

  % Display new/missing/modified methods
  displayMethods(obj,thisClass);

  % Display new/missing interfaces
  %loopValues('Interfaces:',thisClass.getInterfaces);
  diffValues2('Interfaces',thisClass,superClasses,'getInterfaces','',0,obj);

  % Display possible enclosing method
  if isjava(thisClass) && ~isempty(thisClass.getEnclosingMethod)
      disp 'Enclosing method:'
      dispValues('',[],thisClass.getEnclosingMethod)
      disp ' ';
  end

  % Display new/missing constants, sub-classes, constructors etc.
  diffValues2('Sub-classes',  thisClass,superClasses,'getClasses',     '',0,obj);
  diffValues2('Inferior classes', thisClass, [], '', 'InferiorClasses',   0,obj);
  %diffValues2('Constructors',thisClass,superClasses,'getConstructors','',0,obj);
  diffValues2('Annotations',  thisClass,superClasses,'getAnnotations', '',0,obj);
  diffValues2('Events',       thisClass,superClasses,'','EventList',      0,obj);
  diffValues2('Properties',   thisClass,superClasses,'','PropertyList',   0,obj);

  try enums = getEnums(thisClass); catch, enums = {}; end
  loopValues('Constants:',enums);
  diffValues2('Static fields',thisClass,superClasses,'getFields',      '',0,obj,enums);

  % Display class origin (if available)
  displayOrigin(thisClass, thisClassName);

    function loadedClass = loadClass(className)
      try
          loadedClass = java.lang.Class.forName(className);
      catch
          classLoader = com.mathworks.jmi.ClassLoaderManager.getClassLoaderManager; 
          loadedClass = classLoader.loadClass(className);
      end
    end  % loadClass

    function displayHeader()
        disp ' ';
        linkStr = hyperlink(thisClassName,~isClass);
        extraLinks = {};
        if hasUiinspect && isClass
            extraLinks{end+1} = hyperlink('uiinspect',false,'uiinspect',thisClassName);
        end
        if isMcos
            extraLinks{end+1} = hyperlink('doc',false,'doc',thisClassName);
            whichStr = which(thisClassName);
            if ~isempty(whichStr) && strcmp(whichStr(end-1:end),'.m')
                extraLinks{end+1} = hyperlink('help',false,'help',thisClassName);
                extraLinks{end+1} = hyperlink('edit',false,'edit',thisClassName);
            end
        elseif strncmp(thisClassName,'java',4)
            javaVer = char(java.lang.System.getProperty('java.version'));
            if isempty(javaVer)
                javaVer = strtok(regexprep(version('-java'),'Java ',''));
            end
            javaVer = regexprep(javaVer,'\d\.(\d+)\..*','$1');
            urlPath = regexprep(thisClassName,{'\.','\$'},{'/','\.'});
            url = ['https://docs.oracle.com/javase/' javaVer '/docs/api/' urlPath '.html'];
            extraLinks{end+1} = hyperlink('javadoc',false,'web',url);
        end
        if isDotNET
            try
                dotNetVer = char(System.Runtime.InteropServices.RuntimeInformation.FrameworkDescription);
                dotNetVer = regexprep(dotNetVer,'.* (\d+\.+\d+).*','$1');
                if isempty(str2num(dotNetVer)), error('bad value'); end %#ok<ST2NM> 
            catch  % .NET < 4.8
                dotNetVer = System.Environment.Version;
                dotNetVer = [num2str(dotNetVer.Major) '.' num2str(dotNetVer.Minor)];
            end
            url = ['https://docs.microsoft.com/en-us/dotnet/api/' lower(thisClassName) '?view=netframework-' dotNetVer];
            extraLinks{end+1} = hyperlink('MS doc',false,'web',url);
        end
        if ~isempty(extraLinks)
            linkStr = [linkStr ' (' strjoin(extraLinks,', ') ')'];
        end
        if isMcos
            msgStr = [objType ' ' linkStr];
            oldWarn = warning('off','MATLAB:structOnObject');
            modifiers = getMcosModifiers(thisClass);
            warning(oldWarn);
            if ~isempty(modifiers)
                modifiersStr = strrep(modifiers, ',', ', ');
                msgStr = [msgStr ' (' modifiersStr ')'];
            end
            disp(msgStr);
            description = getDescription(thisClass,'',false);
            if ~isempty(description)
                disp(['   ' strtrim(regexprep(description,['^\s*' thisClass.Name],''))]);
            end
            if isClass
                if ~isempty(meta.package.fromName(thisClassName))
                    fprintf('\nNote: %s is also an independent package\n', hyperlink(thisClassName,true))
                end
            else  % package
                if ~isempty(meta.class.fromName(thisClassName))
                    fprintf('\nNote: %s is also an independent class\n', hyperlink(thisClassName,false))
                end
            end
        else  % Java
            modifiers = char(java.lang.reflect.Modifier.toString(thisClass.getModifiers));
            modifiers = strtrim(strrep(modifiers,'public',''));
            if isempty(modifiers),  modifiers = 'Class';  end
            disp([modifiers ' ' linkStr]);
        end
        disp ' ';
    end

    function description = getDescription(object, description, deblank)
        deblank = nargin < 3 || deblank;
        try objectName = object.Name; catch, objectName = char(object); end
        shortName = regexprep(objectName,'.*\.','');
        try
            if nargin < 2 || isempty(description)
                % Fetch the object's description from the MCOS Description fields
                description = '';
                description = strtrim(char(object.Description));
                try description2 = strtrim(char(object.DetailedDescription)); catch, description2 = ''; end
                iscont = ~isempty(description2) && description2(1) > 'Z';  % =lowercase
                if ~isempty(description) && description(end)~='.' && ~iscont, description = [description '.']; end
                try description = strtrim([description ' ' description2]); catch, end
                if deblank, description = regexprep(description,'\s+',' '); end

                % If description not found, extract it from the contents.m file (if available)
                if isempty(description) && isa(object,'meta.class')
                    try
                        filename = which(objectName);
                        folder = fileparts(filename);
                        if ~isempty(folder)
                            filename = fullfile(folder, 'Contents.m');
                            if exist(filename,'file')
                                helpStr = help(filename);
                                if ~isempty(strfind(helpStr,shortName))
                                    description = regexprep(helpStr,['^.*\s' shortName '\s+\-?\s*([^\n]+)\n.*'],'$1');
                                end
                            end
                        end
                    catch
                    end
                end

                % If description still not found, extract it from its help section (if available)
                if isempty(description) && (isa(object,'meta.class') || isa(object,'meta.package'))
                    try
                        helpStr = help(objectName);
                        description = regexprep(helpStr,'\n\s*\n.*','');
                        description = regexprep(description,'\n','\n '); %indent a bit
                        description = regexprep(description,'^\s*(Contents of |methods:).*','','ignorecase');
                        description = regexprep(description,'\s*contained in .*:.*','','ignorecase');
                        if length(helpStr) > length(description) + 5
                            %description = [description ' (' hyperlink('more',false,'help',objectName)];
                        end
                    catch
                    end
                end
            else
                description = char(description);  % not really needed, just in case...
            end

            % Remove extraneous parts from the description
            description = strtrim(description);
            description = regexprep(description, '\.$', ''); %strip trailing '.'
            description = regexprep(description, ['^\s*' shortName  '\s'], '', 'ignorecase'); %strip leading name
            description = regexprep(description, ['^\s*' objectName '\s'], '', 'ignorecase'); %strip leading name
            description = regexprep(description, '^\s*(\-|:)\s*', ''); %strip leading -,:
            description = regexprep(description, '^\s*is (an|a)', '', 'ignorecase'); %strip leading "is a"
            description = regexprep(description, '(Serializable PropInfo)', ' $1');
            description = regexprep(description, 'TODO: Fill in Description', '');
            description = regexprep(description, 'detailed description', '');
            description = regexprep(description, '^\s*\-\- help for.*\-\-', '');
            description = regexprep(description, '^\s*\-\- \w+ not found. Showing.*\-\-', '');
            description = regexprep(description, ['^\s*The ' shortName ' (class|method|function|package|property)$'], '', 'ignorecase');
            description = regexprep(description, '^\s*(class(e)?|method|function|package|property)s?$', '', 'ignorecase');
            description = regexprep(description, '^\s*(description|PropInfo|property)\s*$', '');
            %description = regexprep(description, '^\s*Meta (data|property)', '');
            description = regexprep(description, 'This content is from the documentation of an external language. Some examples and descriptions may not be applicable in MATLAB', '');
            description = regexprep([description ' '], ['[Ss]ee also:? ' thisClassName '\s+'], '');
            description = regexprep(description, '\n+', '\n');
            description = strtrim(description);
            if description(end)=='.' && sum(description=='.')==1, description(end)=''; end
        catch
            % ignore
        end
    end

    function [package,packageName] = getContainingPackage(thisClass,recursive)  %#ok % getContainingPackage
        try
            displayPackage = ~isClass && ~recursive;
            if ~displayPackage
                hContainer = thisClass.ContainingPackage;
            else
                hContainer = thisClass;
                getContainingPackage(thisClass, true)
            end
            if ~isempty(hContainer)
                containerName = hContainer.Name;
                classNames = setdiff({hContainer.ClassList.Name}, thisClass.Name);
                subPackages = {hContainer.PackageList.Name};
                funcNames   = {hContainer.FunctionList.Name};
                numClasses   = numel(classNames);
                numSubPkgs   = numel(subPackages);
                numFunctions = numel(funcNames);
                es = ''; if numClasses + isClass ~= 1, es = 'es'; end
                if numSubPkgs > 0
                    es = sprintf('%s, %d sub-packages', es, numSubPkgs);
                end
                if ~displayPackage
                    pkgStr = hyperlink(containerName,true);
                    pkgType = 'package';
                    if isPython && any(containerName=='.'), pkgType = 'module'; end
                    fprintf('Containing %s (%d class%s):\n', pkgType, numClasses+isClass, es);
                    fprintf('   %s\n\n', pkgStr);
                    return
                end
                % From this point onward, meta.package is handled, not meta.class
                origins = {};
                if numClasses > 0
                    fprintf('Contained classes (%d):\n', numClasses);
                    for classIdx = 1 : numClasses
                        className = classNames{classIdx};
                        baseName = regexprep(className,{'\*(\S)',',(\S)'},{'.$1',', $1'});
                        disp(['   ' hyperlink(baseName,false,linkTarget,className)]);
                        origin = getOriginOf(className);
                        if ~isempty(origin), origins{end+1} = origin; end
                    end
                    fprintf('\n');
                end
                if numSubPkgs > 0
                    fprintf('Contained sub-packages (%d):\n', numSubPkgs)
                    subPackages = sort(subPackages);
                    for pkgIdx = 1 : numSubPkgs
                        disp(['   ' hyperlink(subPackages{pkgIdx},true)]);
                    end
                    fprintf('\n');
                end
                if numFunctions > 0
                    if isPython, pkgType = 'Module'; else, pkgType = 'Package'; end
                    fprintf('%s-level functions (%d):\n', pkgType, numFunctions);
                    [funcNames,sortIdx] = sort(funcNames); %#ok<ASGLU> 
                    hFunctions = hContainer.FunctionList(sortIdx);
                    for funcIdx = 1 : numFunctions
                        %disp(['   ' funcNames{funcIdx}]);
                        thisFunc = hFunctions(funcIdx);
                        inputNames  = getMCOSMethodArgs(thisFunc,'Input');
                        outputNames = getMCOSMethodArgs(thisFunc,'Output');
                        [origin, filename] = getOriginOf([containerName '.' thisFunc.Name]);
                        if ~isempty(origin) && strcmpi(filename(end-1:end),'.m')
                            header = getFileHeader(filename);
                            if isequal(inputNames,{'varargin'})
                                inputArgs = strtrim(strsplit(regexprep(header,'.*=?\s*\((.*)\).*','$1'),','));
                                if ~isempty(inputArgs), inputNames = inputArgs; end
                            end
                            if isequal(outputNames,{'varargout'})
                                outputArgs = strtrim(strsplit(regexprep(header,'(.*)=.*','$1'),','));
                                if ~isempty(outputArgs), outputNames = outputArgs; end
                            end
                        end
                        fprintf('   %s(%s)', thisFunc.Name, strjoin(inputNames,', '));
                        if numel(outputNames) == 1
                            fprintf(' : %s', outputNames{1});
                        elseif numel(outputNames) > 1
                            fprintf(' : [%s]', strjoin(outputNames,', '));
                        end
                        attrs = getMCOSAttributes(thisFunc);
                        if ~isempty(attrs)
                            fprintf('  (%s)', strjoin(attrs,', '));
                        end
                        description = getDescription(thisFunc);
                        if ~isempty(description)
                            fprintf('  %% %s', regexprep(description,'\n.*',''));
                        end
                        fprintf('\n');
                        if ~isempty(origin), origins{end+1} = origin; end
                    end
                    fprintf('\n');
                end
                fprintf('Code origin: ');
                if isPython
                    fprintf('%s', getPythonOrigin(containerName));
                elseif isempty(origins)
                    fprintf('built-in/unknown');
                else
                    origins = unique(origins);
                    for i = 1 : numel(origins)
                        fprintf('%s', parseOrigin(origins{i}));
                    end
                end
                fprintf('\n');
            end
        catch 
            a=1; %#ok<NASGU>  debug breakpoint
        end

        function [origin, filename] = getOriginOf(objectName)
            try
                origin = [];
                filename = which(objectName);
                origin = fileparts(char(filename));
                origin = regexprep(origin,{['^' objectName],'is (an|a)','method$','[\\/]@\w+$'},'');
                origin = strtrim(origin);
            catch
            end
        end
        function header = getFileHeader(origin)
            try
                header = '';
                fid = fopen(origin,'rt');
                txt = fread(fid,'*char')';
                fclose(fid);
                txt = strtrim(regexprep(txt,'%[^\n]*\n',''));
                header = strtrim(regexprep(txt,'function ([^\n]+).*','$1'));
            catch
            end
        end
    end

    function args = getMCOSMethodArgs(thisFunc, argType)
        argNames = thisFunc.([argType 'Names']);
        argTypes = thisFunc.([argType 'Types']);
        if ~iscell(argNames), argNames = {argNames}; end
        if ~iscell(argTypes), argTypes = {argTypes}; end
        argNames = strrep(strcat(argTypes, ' ***', argNames),'***','');
        args = strrep(argNames,'scalar ','');
        %args = regexprep(args,'^any ','string ');
        args = regexprep(args,'^(any)?\s+','');
    end

    function attrs = getMCOSAttributes(thisFunc)
        attrs = {};
        try if ~strcmpi(thisFunc.Access,'public'), attrs = {thisFunc.Access}; end, catch, end
        attrNames = {'Static','Class','Abstract','Sealed','ExplicitConversion','Hidden'};
        for idx = 1 : numel(attrNames)
            thisAttr = attrNames{idx};
            if thisFunc.(thisAttr)
                attrs{end+1} = thisAttr;
            end
        end
    end

    function [superClasses,superclassNames] = getSuperClasses(thisClass,level)
        spacer = '   ';  % initial spacer
        superClasses = thisClass;
        superclassNames = {};
        superclassFound = false;
        disp('Superclasses:');
        try
            if isMcos
                superClasses = thisClass.SuperclassList;
                %superclassFound = ~isempty(superClasses);
                superclassNames = {superClasses.Name};
                %{
                if numel(superclasses) == 1
                    disp(['Superclass: ' hyperlink(superClasses.Name)]);
                else
                    disp('Superclasses:');
                    for classIdx = 1 : numel(superClasses)
                        superclassName = superClasses(classIdx).Name;
                        disp([spacer hyperlink(superclassName)]);
                    end
                end
                %}
                superclassFound = ~isempty(superClasses);
                displayMcosSuperClasses(superClasses,level,spacer);
                if level <= 0 || isinf(level)
                    superClasses = '';
                    superclassNames = '';
                end
            else  % Java
                superclassNames = char(thisClass.getName); % only 1 superclass in Java
                while level>0 && (~superclassFound || level<=inf || ~isempty(strfind(superclassNames,'mathworks'))) % || nargin>1
                    superClasses = superClasses.getSuperclass;
                    if isempty(superClasses)
                        superclassNames = '';
                        break
                    end
                    superclassNames = char(superClasses.getCanonicalName);
                    %disp(['Superclass: hyperlink(superclassName)]);
                    disp([spacer hyperlink(superclassNames)]);
                    superclassFound = true;
                    spacer = [spacer '  '];
                    level = level - 1;
                end
            end
        catch
            % Never mind - maybe no superclass...
        end
        if ~superclassFound
            disp('   (none)');
        end
        %if superclassFound
            disp ' ';
        %end

        function displayMcosSuperClasses(hClasses,level,spacer)
            if level <= 0,  return,  end
            numClasses = numel(hClasses);
            for classIdx2 = 1 : numClasses
                hClass = hClasses(classIdx2);
                className = hClass.Name;
                %str = hyperlink(className);
                baseName = regexprep(className,{'\*(\S)',',(\S)'},{'.$1',', $1'});
                str = hyperlink(baseName,false,linkTarget,className);
                disp([spacer str]);
                if ~any(strncmpi(hClass.Name,{'matlab.graphics.Graphics','matlab.graphics.mixin'},21))
                    try displayMcosSuperClasses(hClass.SuperclassList, level-1, [spacer '  ']); catch, end
                end
            end
        end
    end  % getSuperClasses

    function [objMethods,full] = getMethods(obj,thisClass)
        try
            [objMethods,full] = methods(obj,'-full');
            if isempty(full)
                full = objMethods;
                if isempty(full) && ischar(obj) && nargin > 1
                    hMethods = thisClass.getConstructors();
                    hMethods = [hMethods, thisClass.getMethods()];
                    %className = char(thisClass.getName);
                    for methodIdx = 1 : numel(hMethods)
                        signature = char(hMethods(methodIdx));
                        signature = regexprep(signature, 'public ', '');

                        throwsStr = strtrim(regexprep(signature, '.+(throws .+)', '$1'));
                        if ~strcmp(strtok(throwsStr),'throws'),  throwsStr = '';  end
                        full{methodIdx,6} = throwsStr;
                        signature = regexprep(signature, 'throws .+', '');

                        argsStr = strtrim(regexprep(signature, '.+(\([^)]*\))', '$1'));
                        full{methodIdx,5} = argsStr;
                        signature = regexprep(signature, '\([^)]*\)', '');

                        methodName = strtrim(regexprep(signature, '.+ ([^ ]+)', '$1'));
                        full{methodIdx,4} = methodName;
                        full{methodIdx,3} = regexprep(methodName, '.*\.', '');
                        signature = regexprep(signature, '(.+) [^ ]+', '$1');

                        retVal = strtrim(regexprep(signature, '.+ ([^ ]+)', '$1'));
                        full{methodIdx,2} = retVal;
                        attrs = strtrim(regexprep(signature, '(.+) [^ ]+', '$1'));
                        if ~isequal(attrs, retVal)
                            full{methodIdx,1} = attrs;
                        end
                    end
                else
                    return
                end
            else
                full(:,5) = regexprep(full(:,5), ', ', ',');  % to normalize the params display (no space for Java, space for MCOS)
            end
        catch
            % never mind - maybe no methods
            objMethods = {};
            full = {};
        end
        %full2 = full(:,4:5)'; str = sprintf('%s%s\n',full2{:});
        try full(cellfun('isempty',full(:,5)),5) = {'()'}; catch, end
        try
            if ischar(obj) || isa(obj,'string')
                cls = char(obj);
            else
                try cls = thisClass.Name; catch, cls = char(class(obj)); end
            end
            shortClassName = regexprep(cls,'.*\.','');
            cls = regexprep(cls,{'\$','\.','\*'},{'\\\$','\\\.','\\\*'});
            %full = regexprep(full,cls,shortClassName);
            full(:,4) = regexprep(full(:,4),cls,shortClassName);  % fully-qualified method names
            full(:,[1:3,5]) = regexprep(full(:,[1:3,5]),[cls '(?=(\)|\s|,|$))'],shortClassName);  % attrs,outArgs,methodName,inArgs
            full = regexprep(full,'scalar ','');
        catch
        end
        for methodIdx = 1 : size(full,1)
            methodName = char(full{methodIdx,4});
            methodStr = [methodName, full{methodIdx,5}];
            retVal = char(full{methodIdx,2});
            if ~isempty(retVal) && ~strcmp(retVal,'void') && ~strcmp(retVal,methodName)
                methodStr = [methodStr ' : ' full{methodIdx,2}];
            end
            if ~isempty(full{methodIdx,1})
                methodStr = [methodStr ' (' full{methodIdx,1} ')'];
            end
            objMethods{methodIdx,1} = regexprep(methodStr,'[^(]*\.([^.]+\()','$1','once');
        end
        % Sort the methods lexically
        try
            [dummy, sortIdx] = sort(objMethods(:,1)); %#ok<ASGLU>
            objMethods = objMethods(sortIdx,:);
            full = full(sortIdx,:);
        catch
        end
    end  % getMethods

    function displayMethods(obj, thisClass)
        [objMethods,full] = getMethods(obj,thisClass);
        if ~iscell(superClassNames), superClassNames = {superClassNames}; end
        supMethods = {};
        for superIdx = 1 : numel(superClassNames)
            supMethods = [supMethods; getMethods(superClassNames{superIdx})];
        end
        %if superclassFoundFlag
        newMethods = diffValues('Methods',thisClassName,objMethods,superClassNames,supMethods,1);
        %else
        %    newMethods = {};
        %end
        %diffValues('Methods',thisClassName,obj.methods,superclassName,methods(superclassName),1);
        %diffValues2('Meths',thisClass,superclass,'getMethods');

        % Display modified methods, based on column 4 class
        if ~isempty(full) %&& size(full,2)>=4
            className = [thisClassName,'.'];
            shortClassName = regexprep(className, '^.*\.([^\.]+\.)$', '$1');
            try
                thisClassMethodsIdx = ...
                    strncmp(thisClassName, full(:,4),length(thisClassName)) | ...
                    strncmp(shortClassName,full(:,4),length(shortClassName)) | ...
                    strcmp(shortClassName(1:end-1),full(:,4)); % constructor
            catch
                thisClassMethodsIdx = 1 : size(full,1);
            end
            newMethods  = strtrim(regexprep(newMethods, '\(.*',''));
            objMethods  = objMethods(thisClassMethodsIdx);
            objMethods2 = strtrim(regexprep(objMethods, '\(.*',''));
            [modifiedMethods,idx2] = setdiff(objMethods2,newMethods);
            %modifiedMethods = regexprep(modifiedMethods,[thisClassName '\.'],'');  % remove className prefix
            if ~isempty(modifiedMethods)
                str = 'defined by ';
                if superclassFoundFlag
                    str = 'modified by ';
                end
                disp(['Methods ' str regexprep(thisClassName, '.*\.','') ':']);
                modifiedMethods = objMethods(idx2);
                dispValues('', [], modifiedMethods, obj, supMethods, true)
                disp ' ';
            end
        end
    end

    function enums = getEnums(thisClass)
        try
            % Java objects
            enums = thisClass.getEnumConstants;
        catch
            % MCOS
            enums = {thisClass.EnumerationMemberList.Name};
        end
    end  % getEnums

    function cellStr = toChar(javaObjArray)
      if isempty(javaObjArray)
          cellStr = '';
      else
          try
              cellStr = cellfun(@(c)char(toString(c.getName)),javaObjArray.cell,'un',0);
          catch
              cellStr = cellfun(@char,javaObjArray.cell,'un',0);  % Java annotations start with @ (*.$name)
          end
          cellStr = sort(cellStr);
      end
    end  % toChar

    function loopValues(title,javaObjArray)
        try
            data = toChar(javaObjArray);
        catch
            data = cellfun(@char,cell(javaObjArray),'un',0);
        end
        if ~isempty(data) && iscell(data)
            disp(title);
            maxNameLen = max(cellfun(@length,data));
            itemClass = '';
            for idx = 1 : length(data)
                itemName = data{idx};
                fprintf(['   %-' num2str(maxNameLen) 's'], itemName);
                try
                    itemFQN = [thisClassName '.' itemName];
                    try itemClass = class(itemFQN); catch, end
                    try
                        try
                            itemStr = evalc(['underlyingValue(' itemFQN ')']);
                        catch
                            itemStr = evalc([itemFQN '.get']);
                        end
                    catch
                        try
                            itemStr = evalc([unifiedSuperNames '(' itemFQN ')']);
                        catch
                            itemStr = evalc('cast(eval(itemFQN), itemClass)');
                        end
                    end
                    itemStr = strtrim(itemStr);
                    itemStr = regexprep(itemStr,{'ans =\s*','\n\s*','<[^>]+>','double','^(.+) (.+)$'},{'',' ','','','$1($2)'});
                    basicClassName = regexprep(thisClassName,'.+\.([^\.]+)$','$1');
                    itemStr = regexprep(itemStr,[basicClassName ' enumeration\(' itemName '\)'],'');
                    if ~isempty(itemStr)
                        fprintf(' = %s',itemStr);
                    end
                catch
                end
                fprintf('\n');
            end
            fprintf('\n');
            if ~isempty(data) && ~isempty(itemClass)
                fprintf('Underlying class: %s\n\n', itemClass);
            end
        elseif isMcos
            try allProperties = thisClass.PropertyList;       catch, return, end
            try constantIdx = find([allProperties.Constant]); catch, return, end
            if isempty(constantIdx), return, end
            disp(title);
            propNames = {allProperties(constantIdx).Name};
            maxNameLen = max(cellfun(@length,propNames));
            thisObj = obj; if ischar(obj), thisObj = eval(obj); end
            for idx = 1 : numel(propNames)
                propName = propNames{idx};
                try
                    try
                        thisValue = thisObj.(propName);
                    catch
                        propDef = allProperties(constantIdx(idx));
                        thisValue = propDef.DefaultValue;
                    end
                    extraStr = [' : ' class(thisValue) ...
                                ' = ' charizeData(thisValue)];
                catch
                    extraStr = '';
                end
                fprintf(['   %-' num2str(maxNameLen) 's%s\n'], propName, extraStr);
            end
            fprintf('\n');
        end
    end  % loopValues

    function values = diffValues(title,thisClassName,thisValues,superClassNames,superValues, flag, obj, enums)
        if nargin < 7, obj = []; end
        try thisValues = setdiff(toChar(thisValues), toChar(enums)); catch, end
        if isjava(thisValues),   thisValues  = toChar(thisValues);   end
        if isjava(superValues),  superValues = toChar(superValues);  end
        if isempty(thisValues),  thisValues = '';  end
        regex1 = {' [(].*', ' :.+', ' % .*', '([\(,])[^,\)]+', '\(_\)'};  %, '([\(,])\s*(\w+)[^,\)]*'
        regex2 = {'',       '',     '',      '$1_',            ''};       %, '$1$2'
        thisValuesSimple = regexprep(thisValues, regex1, regex2);
        try
            superValuesSimple = regexprep(superValues, regex1, regex2);
        catch
            superValuesSimple = '';
        end
        thisClassName2  = regexprep(thisClassName, '[^,]*\.','');
        try superClassNames = strjoin(superClassNames,','); catch, end
        superClassName2 = regexprep(superClassNames,'[^,]*\.','');
        if nargin >= 6 && flag
            thisValues2  = regexprep(thisValuesSimple,  ['^' thisClassName2], superClassName2);  % =stripValue(thisValues, thisClassName2);
            superValues2 = regexprep(superValuesSimple, ['^' superClassName2], thisClassName2);  % =stripValue(superValues,superClassName2);
            superNames = strsplit(superClassName2,',');
            for idx = 1 : numel(superNames)
                thisSuperName = superNames{idx};
                % Special handling for the <handle> class
                if ~strcmp(thisSuperName,'handle')
                    superValues2 = regexprep(superValues2,[thisSuperName '(?=(\s|\)|,))'],thisClassName2);
                end
            end
            % Remove superclasses' constructors
            constructorsIdx = ~cellfun('isempty',regexp(superValues2,['^\s*' thisClassName2 '\(']));
            superFuncNames = regexprep(superValues2,{'\(.*',' *:.*'},'');
            for idx = 1 : numel(superNames)
                thisSuperName = superNames{idx};
                constructorsIdx = constructorsIdx | strcmp(thisSuperName,superFuncNames);
            end
            superValues (constructorsIdx) = [];
            superValues2(constructorsIdx) = [];
            superValuesSimple(constructorsIdx) = [];
        else
            thisValues2  = thisValuesSimple;
            superValues2 = superValuesSimple;
        end
        [values,idx] = getNewValues(superValuesSimple,thisValues2); %,superValues2);
        if ~isempty(values) && ~isempty(superValues)
            str = [title ' in ' unifiedSuperNames];
            if ~isempty(unifiedSuperNames) && ~strcmp(thisClassName2,unifiedSuperNames)
                str = [str ' missing in ' thisClassName2];
            end
            disp([str ':'])
            dispValues(title, superClassNames, unique(superValues(idx)), obj, thisValues, false)
            disp ' ';
        end
        [values,idx] = getNewValues(thisValuesSimple,superValues2); %,thisValuesSimple);
        if strcmpi(title,'Inferior classes')
            for i = 1 : numel(thisValues)
                [className,rest] = strtok(thisValues{i});
                thisValues{1} = [hyperlink(className) rest];
            end
        end
        if ~isempty(values) && ~isempty(thisValues)
            str = title;
            if ~isempty(superClassNames), str = [str ' in ' thisClassName2]; end
            if ~isempty(superClassNames) && ~strcmp(thisClassName2,unifiedSuperNames)
                str = [str ' missing in ' unifiedSuperNames];
            end
            disp([str ':'])
            dispValues(title, thisClassName, unique(thisValues(idx)), obj, [])
            disp ' ';
        end
    end  % diffValues

    function [newValues,idx] = getNewValues(set1, set2)  % vals in set1 that are not in set2
        % setdiff() removes duplicates, so we use ismember() instead
        %[newValues,idx] = setdiff(set1,set2);
        %{
        newValues = set1;
        [idx2,idx1] = ismember(set2,set1b); %#ok<ASGLU>
        idx1(idx1<=0) = [];
        newValues(idx1) = [];
        idx = setdiff(1:numel(set1), idx1);
        %}
        idx = 1 : numel(set1);
        set2 = unique(set2);
        for idx2 = 1 : numel(set2)
            idx1 = ismember(set1, set2{idx2});
            idx(idx1) = 0;
        end
        idx(idx<=0) = [];
        newValues = set1(idx);
    end

    function dispValues(title,classname,values,obj,otherValues,dispOrSkip)
        oldWarn = warning('off','MATLAB:namelengthmaxexceeded');
        try
            try
                valuesStr = '';
                staticFlag = strcmpi(strtok(title),'static');
                maxFieldLen = max(cellfun(@length,values));
                try allMethodNames = {thisClass.MethodList.Name}; catch, end
                valuesSimple = regexprep(values,{'\(.*',' *:.*'},'');
                try otherValuesSimple = regexprep(otherValues,{'\(.*',' *:.*'},''); catch, end
                for idx = 1 : length(values)
                    thisValue = values{idx};
                    try
                        if staticFlag
                            try
                                dataValue = eval([classname '.' thisValue]);
                            catch
                                dataValue = obj.(thisValue); %=eval(['obj.' thisValue]);
                            end
                            if isa(dataValue,'java.lang.String')
                                dataValue = ['''' char(dataValue) ''''];
                            elseif isa(dataValue,'java.lang.String[]')
                                dataValue = cell(dataValue);
                                dataValue = sprintf('''%s'',', dataValue{:});
                                dataValue = ['{' dataValue(1:end-1) '}'];
                            end
                            padStr = repmat(' ',1,maxFieldLen-length(thisValue));
                            try
                                valStr = num2str(dataValue);
                            catch
                                valStr = char(dataValue);
                            end
                            newStr = sprintf('   %s%s = %s',num2str(thisValue),padStr,valStr);
                        else
                            newStr = sprintf('   %s',num2str(thisValue));
                        end
                    catch
                        newStr = sprintf('   %s',num2str(thisValue));
                    end
                    newStr = regexprep(newStr, '\(bytes=.+\)', '(bytes=...)');
                    extraStr = '';
                    thisValue = valuesSimple{idx};
                    if nargin >= 5 && ~isempty(otherValues)
                        % Compare to superclasses definition
                        otherIdx = find(strcmpi(thisValue,otherValuesSimple),1);
                        if ~isempty(otherIdx)
                            otherDef = otherValues{otherIdx};
                            if nargin < 6 || dispOrSkip
                                if ~strcmpi(strrep(newStr,' ',''), strrep(otherDef,' ',''))
                                   newStr2 = regexprep(newStr,{'^\s*','\(.*','\s*:.*'},'');
                                   if ~strcmpi(newStr2, strrep(otherDef,' ',''))
                                       extraStr = ['  % [superclasses: ' otherDef ']'];
                                   end
                                end
                            else
                                continue  % skip redefined value
                            end
                        end
                    end
                    % Display methods description (if available)
                    if nargin >= 5 && isa(thisClass,'meta.class')
                        try
                            methodIdx = find(strcmp(thisValue,allMethodNames),1);
                            if ~isempty(methodIdx)
                                thisMethod = thisClass.MethodList(methodIdx);
                                attrs = getMCOSAttributes(thisMethod);
                                if ~isempty(attrs)
                                    newStr = regexprep(newStr,' \(.+\)$',''); %replaced with attrs below
                                    extraStr = sprintf('  (%s)%s', strjoin(attrs,', '), extraStr);
                                end
                                description = getDescription(thisMethod);
                                if ~isempty(description)
                                    if ~any(extraStr=='%'), extraStr = [extraStr '  %']; end
                                    extraStr = sprintf('%s %s', extraStr, regexprep(description,'\n.*',''));
                                end
                            end
                        catch
                        end
                    end
                    newStr = [newStr extraStr];
                    valuesStr = [valuesStr newStr newline];
                end
                values = valuesStr;
            catch
                values = regexprep(evalc('disp(values)'), ' ''([^\n]*)''', '$1');
            end
            values = regexprep(values, {', ', ','}, {',', ', '}); %=strrep(values, ',', ', ');  % ','->', ' except if already has a space after
            values(values=='*') = '.';  % Yair 7/7/17 - some .Net classes are reported as A*B*C
            values = regexprep(values, '([\w]+(\.[\w$]+)+)', hyperlink('$1'));
            values = regexprep(values, '<a href=[^>]+>([\d\.-]+)</a>', '$1'); %de-hyperlink numbers
            values = regexprep(values, '<a href=[^>]+>(\w+\.\d+)</a>', '$1'); %de-hyperlink xyz.##
            values = regexprep(values, '<a href=[^>]+>([Ii]\.e|[Ee]\.g)</a>','$1'); %de-hyperlink i.e., e.g.
            if values(end)==10,  values(end)=[];  end
        catch
            % never mind...
            a=1;  %#ok debug breakpoint
        end
        warning(oldWarn);
        disp(values);
    end  % dispValues

    %{
    function cellStr = stripValue(cellStr,value)  %#ok unused
      if ~isempty(cellStr)
          value = regexprep(value,'.*\.','');
          cellStr = setdiff(cellStr, value);
          cellStr(~cellfun('isempty',regexp(cellStr,['^',value]))) = [];  % strip constructors
      end
    end  % stripValue
    %}

    function diffValues2(title,thisClass,superClasses,javaOpName,mcosOpName,varargin)
        try
            if isjava(thisClass) && ~isempty(javaOpName)
                try
                    % Java only allows a single superclass
                    diffValues(title, char(thisClass.getName),    awtinvoke(thisClass,javaOpName), ...
                                      char(superClasses.getName), awtinvoke(superClasses,javaOpName), varargin{:});
                catch
                    % Never mind - maybe no superclass...
                    diffValues(title, char(thisClass.getName),    awtinvoke(thisClass,javaOpName), '', {}, varargin{:});
                end
            elseif ~isjava(thisClass) && ~isempty(mcosOpName)  % MCOS
                valuesSuper = {};
                values = getMcosFields(thisClass,mcosOpName);
                if strcmpi(title,'Properties')
                    thisObj = obj;
                    if ischar(obj)
                        try
                            thisObj = eval(obj);
                        catch
                            thisObj = meta.class.fromName(obj);
                        end
                    end
                    if isempty(values) && ~isequal(thisClass,thisObj) % && isPython
                        try
                            values = fieldnames(thisObj);
                            for idx = 1 : numel(values)
                                propName = values{idx};
                                thisValue = thisObj.(propName);
                                values{idx} = [propName ' : ' class(thisValue) ...
                                                        ' = ' charizeData(thisValue)];
                            end
                        catch
                        end
                    end
                    if ~isempty(values) && isMcos
                        % Exclude constant properties (separate treatment with enums)
                        for idx = numel(values) : -1 : 1
                            try
                                propName = strtok(values{idx});
                                if findprop(thisObj,propName).Constant
                                    values(idx) = [];
                                end
                            catch
                            end
                        end
                    end
                end
                superNames = '';
                if ~isempty(superClasses) && ~isequal(thisClass,superClasses)
                    % Matlab allows multiple superclasses
                    for idx = 1 : numel(superClasses)
                        valuesSuper = [valuesSuper,getMcosFields(superClasses(idx),mcosOpName)];
                    end
                    superNames = superClassNames;
                end
                diffValues(title, char(thisClass.Name), values, superNames, valuesSuper, varargin{:});
            end
        catch e %#ok<NASGU>
            % ignore
        end
    end  % diffValues2

    function values = getMcosFields(thisClass,mcosFieldName)
        values = thisClass.(mcosFieldName);
        %try values = {values.Name}; catch, end
        if ~isempty(values)
            valueCells = {};
            isCells = iscell(values);
            oldWarn = warning('off','MATLAB:structOnObject');
            for idx = 1 : numel(values)
                try
                    if isCells %try
                        thisValue = values{idx};
                    else %catch
                        thisValue = values(idx);
                    end
                    valueCells{idx} = thisValue.Name;

                    % Append data type (if available)
                    try type = char(thisValue.Type.Name); catch, type = ''; end
                    if ~isempty(type) && ~strcmpi(type,'any')
                        type = strrep(type, 'unicodeString', 'char');
                        type = strrep(type, [thisValue.DefiningClass.Name ': '], '');
                        type = strrep(type, ', ', ',');
                        valueCells{idx} = [valueCells{idx} ' : ' type];
                    end

                    % Append non-default modifiers (if any)
                    modifiers = getMcosModifiers(thisValue);
                    if ~isempty(modifiers)
                        modifiers = [' (' modifiers ')'];
                        valueCells{idx} = [valueCells{idx} ' ' modifiers];
                    end

                    % Append description (if available)
                    description = getDescription(thisValue);
                    hasDescription = ~isempty(description); % && ~any(strcmpi(description,{'description','PropInfo','property'}));
                    if hasDescription
                        description = regexprep(description,'\n.*','');  % display only top description line
                        valueCells{idx} = [valueCells{idx} '  % ' description];
                    end
                catch err %#ok<NASGU>
                    % never mind
                end
            end
            warning(oldWarn);
            values = valueCells;
        else
            values = {};  % not: [0x1 SomeClass]
        end
    end  % getMcosFields

    function modifiersStr = getMcosModifiers(thisClass)
        modifiersStr = ' ';
        thisClassStruct = struct(thisClass);
        fns = fieldnames(thisClassStruct);
        hasGetAccessField = any(strcmpi(fns,'GetAccess'));
        delDefault = false;
        for fieldIdx = 1 : numel(fns)
            thisFieldName = fns{fieldIdx};
            thisFieldValue = thisClassStruct.(thisFieldName);
            if iscell(thisFieldValue)
                if isempty(thisFieldValue)
                    thisFieldValue = '{}';
                else
                    try
                        fieldClass = class([thisFieldValue{:}]);
                    catch
                        fieldClass = 'elements';
                    end
                    numVals = numel(thisFieldValue);
                    if numVals == 1 && strcmpi(fieldClass,'meta.class')
                        thisFieldValue = thisFieldValue{1}.Name;
                    else
                        thisFieldValue = sprintf('{%d %s}',numVals,fieldClass);
                    end
                end
            end
            if isempty(thisFieldValue)
                continue
            elseif strcmpi(thisFieldName,'Access') && hasGetAccessField
                continue  % ignore redundant Access field
            elseif ~isempty(strfind(thisFieldName,'Method')) %&& ~isempty(thisFieldValue)
                modifiersStr = [modifiersStr thisFieldName ','];
            elseif strcmpi(thisFieldName,'Validation') %&& ~isempty(thisFieldValue)
                %modifiersStr = [modifiersStr 'Validation,'];
                subFields = fieldnames(thisFieldValue);
                subFieldCells = {};
                for idx = 1 : numel(subFields)
                    thisSubField = subFields{idx};
                    if ~isempty(thisFieldValue.(thisSubField))
                        subFieldCells{end+1} = thisSubField;
                    end
                end
                validationStr = strjoin(subFieldCells,',');
                modifiersStr = [modifiersStr 'Validation=[' validationStr '],'];
            elseif strcmpi(thisFieldName,'HasDefault') && thisFieldValue %==true
                try
                    defaultValue = thisClassStruct.DefaultValue;
                    defaultValueStr = charizeData(defaultValue);
                    if ~isequal(defaultValue,[]) %~strcmp(defaultValue,'[]')
                        modifiersStr = [modifiersStr 'DefaultValue=' defaultValueStr ','];
                    end
                    delDefault = true;
                catch
                    modifiersStr = [modifiersStr 'HasDefault,'];
                end
            elseif islogical(thisFieldValue) && thisFieldValue %%isequal(thisFieldValue,true)
                modifiersStr = [modifiersStr thisFieldName ','];
            elseif strcmpi(thisFieldName,'PartialMatchPriority') && thisFieldValue ~= 1
                modifiersStr = [modifiersStr thisFieldName '=' num2str(thisFieldValue) ','];
            elseif ~isempty(strfind(thisFieldName,'Access')) && ~isequal(thisFieldValue,'public')
                modifiersStr = [modifiersStr thisFieldName '=' thisFieldValue ','];
            end
            if delDefault
                modifiersStr = strrep(modifiersStr,'DefaultValue,','');
            end
        end
        modifiersStr = strtrim(modifiersStr(1:end-1));
        modifiersStr = regexprep(modifiersStr,'GetAccess=(.+),SetAccess=\1','Access=$1');
    end  % getMcosModifiers

    function data = charizeData(data)
        if isa(data, 'com.mathworks.hg.types.HGCallback')
            data = get(data, 'Callback');
        end
        if isempty(data)
            if ischar(data)
                data = '''''';
            elseif iscell(data)
                data = '{}';
            elseif isnumeric(data) && strcmpi(class(data),'double')
                data = '[]';
            else
                sizeStr = regexprep(num2str(size(data)),'\s+',char(215)); %char(215)='×' ('x')
                data = ['[' sizeStr ' ' char(class(data)) ']'];
            end
        elseif isa(data,'py.str')
            data = ['''' char(data) ''''];
        elseif ~ischar(data) && ~isa(data, 'java.lang.String')
            try newData = mat2str(data); catch, newData = ''; end
            if isempty(newData) || any(newData==';')
                newData = strtrim(evalc('disp(data)'));
            end
            try
                % multi-column arrays
                newData = regexprep(newData, '  +', ' ');
                newData = regexprep(newData, 'Columns \d+ through \d+\s', '');
                newData = regexprep(newData, 'Column \d+\s', '');

                % Python objects
                newData = regexprep(newData, 'Python <[^<]+</a>,? ?', '');
                newData = regexprep(newData, '^(\s+)?with no properties\.?,? ?', '');
                newData = regexprep(newData, '^(\s+)?with properties:.+]\s*', '');
            catch
            end
            if iscell(data)
                newData = ['{ ' newData ' }'];
            elseif isempty(data)
                newData = '';
            elseif any(ishandle(data)) || ((isnumeric(data) || islogical(data)) && numel(data) > 1)
                newData = ['[' newData ']'];
                newData = regexprep(newData, '\[(\d+)\]$', '$1');  %'[0]' => '0'
            end
            newData = regexprep(newData,{'<a href=\S+ ([^"]+).*</a>',' ?with properties:\s+(.*)','\n\s+',':,'},{'$1','<$1>',',',': '});
            newData = regexprep(newData,'\s+',' ');
            newData = regexprep(newData,'^,','');
            if any(newData==' '), newData = ['[' newData ']']; end
            data = newData;
        elseif ischar(data) %~isempty(data)
            data = ['''' data ''''];
        else %string
            data = ['"' data '"'];
        end
    end

    function str = hyperlink(base,isPackage,linkOp,uri)
        if any(strcmpi(base,{'e.g','i.e'})), str=base; return, end
        if nargin < 3,  linkOp = linkTarget;  end
        if nargin < 4,  uri = base;  end
        str = ['<a href="matlab:' linkOp '(''' uri ''''];
        if nargin > 1 && isPackage
            str = [str ',1,''package'''];
        end
        str = [str ')">' base '</a>'];
    end

    function displayOrigin(thisClass, thisClassName)
        % Display class origin information: http://undocumentedmatlab.com/blog/static-java-classpath-hacks
        try  % Java class
            whereisjavaclassloadingfrom(thisClassName);
        catch  % MCOS/other class
            if isDotNET
                try
                    if isempty(dotNetVer), dotNetVer = '(unknown)'; end
                    disp(['.NET Framework: ' dotNetVer]);
                    try
                        hDotNetType = System.Type.GetType(thisClassName,false);
                        hAssembly = hDotNetType.Assembly;
                    catch
                        hAssembly = System.Reflection.Assembly.GetExecutingAssembly;
                    end
                    disp(['Assembly path:  ' char(hAssembly.Location)]);
                    disp(['Assembly desc:  ' char(hAssembly.ToString)]);
                    fprintf('\n');
                catch
                end
            end
            if isPython
                origin = getPythonOrigin(origObj);
            else
                origin = char(which(thisClassName));
            end
            if ~isempty(origin) && (~isDotNET || isempty(strfind(origin,'built')))
                origin = getDescription(thisClass, origin);
                origin = parseOrigin(origin);
                fprintf('Code origin: %s\n\n', origin);
            end
        end
    end
    
    function origin = getPythonOrigin(object)
        try
            if ischar(object) || isa(object,'string')
                object = char(object);
                object = py.importlib.import_module(object(4:end)); %strip leading 'py.'
            end
            origin = parseOrigin(char(py.inspect.getfile(object)));
        catch
            origin = 'built-in/unknown';
        end
    end

    function origin = parseOrigin(origin)
        unknownStr = 'built-in/unknown';
        if isempty(origin)
            origin = unknownStr;
        else
            origin = regexprep(origin, 'built.?in method',unknownStr);
        end
        if ~strncmpi(origin,'built',5) && isempty(strfind(origin,unknownStr))
            [fpath,fname,fext] = fileparts(origin);
            if any(strcmpi(fext,{'.m','.mlx'}))  % edit origin as a MAtlab file
                origin = hyperlink(origin,false,'edit');
            elseif ispc && strcmpi(fext,'.p') % open parent folder (Windows only)
                origin = [hyperlink(fpath,false,'winopen') '\' fname fext];
            elseif ispc % open entire origin as a folder (Windows only)
                origin = hyperlink(origin,false,'winopen');
            end
            origin = sprintf('\n   %s',origin);
        end
    end

    function whereisjavaclassloadingfrom(ClassName)
        %WHEREISJAVACLASSLOADINGFROM  Show where a Java class is loaded from
        %
        % whereisjavaclassloadingfrom(ClassName)
        %
        % Shows where a Java class is loaded from in this Matlab session's JVM.
        % This is for diagnosing Java class load problems, such as classpath
        % ordering issues, seeing if a class of a given name is included in an
        % unexpected JAR file, etc.
        %
        % Displays output to console.
        %
        % Examples:
        %
        % whereisjavaclassloadingfrom('java.util.HashMap')
        % whereisjavaclassloadingfrom('com.ldhenergy.etools.MxUtil')
        % whereisjavaclassloadingfrom('com.google.common.collect.Maps')
        % whereisjavaclassloadingfrom('org.apache.commons.math.complex.Complex')
        
        % (c) Andrew Janke http://stackoverflow.com/questions/4376565/java-jpa-class-for-matlab#4380622
        % http://undocumentedmatlab.com/blog/static-java-classpath-hacks#analyze

        % Use javaArray to get Class object without having to instantiate. This
        % lets it work with objects that have private or non-zero-arg constructors,
        % and avoids side effects of object construction.
        % (Would use java.lang.Class.forName(), because that's a more direct way of
        % doing this, but it doesn't work for stuff on the dynamic classpath.)
        ja = javaArray(ClassName,1);
        klass = ja.getClass().getComponentType();

        klassLoader = klass.getClassLoader();
        if isempty(klassLoader)
            % JVM used null to represent the "bootstrap" class loader
            % I think that's the same as the "system" class loader
            klassLoader = java.lang.ClassLoader.getSystemClassLoader();
        end
        klassLoaderStr = char(klassLoader);

        klassFilePath = [strrep(ClassName, '.', '/') '.class'];
        try
            % This logic assumes that the classes exist as files in the class
            % loader. It's a valid assumption for mainstream class loaders,
            % including the one's I've seen with Matlab.
            klassUrl = klassLoader.getResource(klassFilePath);
            if isempty(klassUrl)
                klassUrlStr = klassFilePath; %='';
            else
                klassUrlStr = char(klassUrl);
            end
        catch err
            klassUrlStr = sprintf('ERROR: %s', err.message);
        end

        % Get all locations, to reveal masked definitions
        urls = enumeration2array(klassLoader.getResources(klassFilePath));

        fprintf('ClassLoader: %s\n', klassLoaderStr);
        if ~isempty(strfind(klassUrlStr,'file:'))
            filename = regexprep(klassUrlStr,'.*file:/?/?','');
            filename = regexprep(filename,'%20',' ');
            filename = regexprep(filename,'!.*','');
            filename = strrep(filename,'/','\');
            if ispc, filename = hyperlink(filename,false,'winopen'); end
            fprintf('File: %s\n',filename);
        else
            fprintf('File: unknown!\n');
        end
        fprintf('URL: %s\n', klassUrlStr);
        if numel(urls) > 1
            %disp('Masked (shaddowed):');
            for i = 2:numel(urls)
                fprintf('URL: %s (masked/shaddowed)\n', char(urls(i)));
            end
        end
        fprintf('\n');

        function out = enumeration2array(jenum)
            tmp = {};
            while jenum.hasMoreElements()
                tmp{end+1} = jenum.nextElement();
            end
            out = [tmp{:}];
        end
    end  % whereisjavaclassloadingfrom

end  % checkClass
